# -*- coding: utf-8 -*-
#!/usr/bin/env python3

import asyncio
import base64
import json
import os
import sys

from aiohttp import web
from asyncdb import AsyncDB

from navigator.conf import DEBUG, SESSION_PREFIX, SESSION_URL, asyncpg_url, config
from navigator.handlers import nav_exception_handler
from navigator.modules.session.session import AbstractSession
from navigator.views import BaseView


class UserSession(BaseView):
    async def get(self):
        try:
            session = self.request["session"]
        except KeyError:
            session = None
        try:
            if not session:
                headers = {"x-status": "Empty", "x-message": "Invalid User Session"}
                return self.no_content(headers=headers)
            else:
                headers = {"x-status": "OK", "x-message": "Session OK"}
                data = session.content()
                if data:
                    return self.json_response(response=data, headers=headers)
        except Exception as err:
            return self.error(request, exception=err)


class navSession(object):
    """
    navSession
       Basic Interaction with Session backend of Django
    """

    _loop = None
    _redis = None
    _cache = None
    _result = {}
    _session = None

    def __init__(self, dsn="", session: AbstractSession = None, loop=None):
        if loop:
            self._loop = loop
        else:
            self._loop = asyncio.get_event_loop()
        self._result = {}
        self._session_key = ""
        self._session_id = None
        self._loop.set_exception_handler(nav_exception_handler)
        # TODO: define other session backend
        self._cache = redis = AsyncDB("redis", dsn=dsn)
        self._session = session

    def cache(self):
        return self._cache

    def set_result(self, result):
        self._result = result

    async def connect(self):
        await self._cache.connection()
        if self._cache:
            self._session.Backend(self)

    async def close(self):
        if self._cache:
            await self._cache.close()
            await self._cache.wait_closed()

    async def logout(self):
        """
        logout.
           Clear all session info
        """
        self._session_id = None
        self._session_key = None
        self._result = {}

    async def decode(self, key):
        return await self._session.decode(key)

    async def encode(self, key, data):
        return await self._session.encode(key, data)

    def session_key(self):
        return self._session_key

    def session_id(self):
        return self._session_id

    def id(self, id):
        self._session_id = id

    def content(self):
        return self._result

    async def get(self, key):
        if key in self._result:
            return self._result[key]
        else:
            return None

    async def set(self, key, value):
        self._result[key] = value

    async def add(self, **kwargs):
        self._result = {**self._result, **kwargs}

    """
    Magic Methods
    """

    def __getitem__(self, key):
        """
        Sequence-like operators
        """
        if self._result:
            return self._result[key]
        else:
            return False

    def __contains__(self, key):
        if key in self._result:
            return True
        else:
            return False

    def __iter__(self):
        return iter(self._result)

    def __getattr__(self, key):
        if self._result:
            return self._result[key]
        else:
            raise KeyError("Invalid Session")



#!/usr/bin/env python3
import asyncio
import importlib
import inspect

# logging system
import logging
import os
import sys
import typing
from abc import ABC, abstractmethod
from logging.config import dictConfig

# from aiohttp_swagger import setup_swagger
from pathlib import Path
from typing import Any, Callable, Dict, List

import aiohttp_cors
import aiohttp_jinja2
import jinja2
from aiohttp import web
from aiohttp.abc import AbstractView
from aiohttp.web import middleware
from aiojobs.aiohttp import setup, spawn
from asyncdb import AsyncPool
from asyncdb.providers.redis import redis

from navigator.conf import (
    API_NAME,
    APP_DIR,
    BASE_DIR,
    DEBUG,
    INSTALLED_APPS,
    STATIC_DIR,
    logdir,
)
from navigator.middlewares import basic_middleware

# make a home and a ping class
from navigator.resources import home, ping

loglevel = logging.INFO

logger_config = dict(
    version=1,
    formatters={
        "console": {"format": "%(message)s"},
        "file": {
            "format": "%(asctime)s: [%(levelname)s]: %(pathname)s: %(lineno)d: \n%(message)s\n"
        },
        "default": {"format": "[%(levelname)s] %(asctime)s %(name)s: %(message)s"},
    },
    handlers={
        "console": {
            "formatter": "console",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
            "level": loglevel,
        },
        "StreamHandler": {
            "class": "logging.StreamHandler",
            "formatter": "default",
            "level": loglevel,
        },
    },
    root={
        "handlers": ["StreamHandler"],
        "level": loglevel,
    },
)
dictConfig(logger_config)

#######################
##
## PATH CONFIGURATION
##
#######################


class path(object):
    """
    path.
    description: django-like URL router configuration
    """

    method = ""
    url = ""
    handler = None
    name = ""

    def __init__(self, method, url, handler, name=""):
        self.method = method
        self.url = url
        self.handler = handler
        self.name = name


#######################
##
## APPS CONFIGURATION
##
#######################


def app_startup(app_list: list, app: web.Application, context: dict, **kwargs: dict):
    # Configure the main App
    app.router.add_route("GET", "/ping", ping)
    # index
    app.router.add_get("/", home)
    for app_name in app_list:
        obj = None
        try:
            name = app_name.split(".")[1]
            app_class = importlib.import_module(app_name, package="apps")
            obj = getattr(app_class, name)
            instance = obj(context, **kwargs)
            app.add_subapp("/{}/".format(name), instance.App)
            # TODO: build automatic documentation
        except ImportError as err:
            print(err)
            continue


class AppHandler(ABC):
    """
    AppHandler.

    Main Class for registration of Callbacks, Signals, Route Initialization, etc
     * TODO: adding support for middlewares
     * TODO: get APP names
    """

    _name = None
    logger = None
    _loop = None
    debug = False
    app: web.Application = None
    __version__: str = "0.0.1"
    app_description: str = ""
    cors = None
    _middleware: Any = None
    auto_home: bool = True
    enable_aiojobs: bool = False
    enable_static: bool = True
    staticdir: str = ""

    def __init__(self, context: dict, *args: List, **kwargs: dict):
        self._name = type(self).__name__
        self.logger = logging.getLogger(self._name)
        # configuring asyncio loop
        self._loop = asyncio.get_event_loop()
        self.app = self.CreateApp()
        # config
        self.app["config"] = context
        # register signals for startup cleanup and shutdown
        self.app.on_startup.append(self.on_startup)
        self.app.on_cleanup.append(self.pre_cleanup)
        self.app.on_cleanup.append(self.on_cleanup)
        self.app.on_shutdown.append(self.on_shutdown)
        self.app.on_response_prepare.append(self.on_prepare)
        # TODO: making automatic discovery of routes
        if self.auto_home:
            self.app.router.add_route("GET", "/ping", ping)
            self.app.router.add_route("GET", "/", home)

    def CreateApp(self) -> web.Application:
        if DEBUG:
            print("SETUP NEW APPLICATION: {}".format(self._name))
        middlewares = {}
        self.cors = None
        if self._middleware:
            middlewares = {"middlewares": self._middleware}
        app = web.Application(
            logger=self.logger,
            client_max_size=(1024 * 1024) * 1024,
            loop=self._loop,
            **middlewares
        )
        self.cors = aiohttp_cors.setup(
            app,
            defaults={
                "*": aiohttp_cors.ResourceOptions(
                    allow_credentials=True,
                    expose_headers="*",
                    allow_methods="*",
                    allow_headers="*",
                    max_age=3600,
                )
            },
        )
        return app

    @property
    def App(self) -> web.Application:
        return self.app

    @property
    def Name(self) -> str:
        return self._name

    def configure(self) -> None:
        """
        configure.
            making configuration of routes
        """
        if self.enable_aiojobs:
            # Adding setup and support for aiojobs
            setup(self.app)
        if self.enable_static:
            # adding statics
            # TODO: can personalize the path
            static = self.staticdir if self.staticdir else STATIC_DIR
            self.app.router.add_static(
                "/static/", path=static, name="static", append_version=True
            )
            # self.app.add_routes(
            #     [web.static('/static', static, append_version=True)]
            # )

    def setup_docs(self) -> None:
        """
        set_cors.
        description: define CORS configuration
        """
        # Configure CORS, swagger and documentation from all routes.
        for route in list(self.app.router.routes()):
            fn = route.handler
            signature = inspect.signature(fn)
            doc = fn.__doc__
            if doc is None:
                # TODO: making more efficiently
                fnName = fn.__name__
                if signature.return_annotation:
                    response = str(signature.return_annotation)
                else:
                    response = "unknown"
                if fnName not in [
                    "_handle",
                    "channel_handler",
                    "WebSocket",
                    "websocket",
                ]:
                    doc = """
                    summary: {fnName}
                    produces:
                    - text/plain
                    responses:
                        "200":
                            description: Successful operation
                            content:
                                {response}""".format(
                        fnName=fnName, response=response
                    )
                    try:
                        fn.__doc__ = doc
                    except (AttributeError, ValueError):
                        pass

    def setup_cors(self, cors):
        for route in list(self.app.router.routes()):
            try:
                # if DEBUG:
                #     self.logger.info(f'Adding CORS to {route.method} {route.handler}')
                # if not isinstance(route.resource, web.StaticResource):
                if inspect.isclass(route.handler) and issubclass(
                    route.handler, AbstractView
                ):
                    cors.add(route, webview=True)
                else:
                    cors.add(route)
            except ValueError as err:
                print("Error on Adding CORS: ", err)
                pass

    async def on_prepare(self, request, response):
        """
        on_prepare.
        description: Signal for customize the response while is prepared.
        """
        pass

    async def pre_cleanup(self, app):
        """
        pre_cleanup.
        description: Signal for customize the response when server is closing
        """
        pass

    async def on_cleanup(self, app):
        """
        on_cleanup.
        description: Signal for customize the response when server is closing
        """
        pass

    async def on_startup(self, app):
        """
        on_startup.
        description: Signal for customize the response when server is started
        """
        pass

    async def on_shutdown(self, app):
        """
        on_shutdown.
        description: Signal for customize the response when server is shutting down
        """
        pass


class AppBase(AppHandler):
    _middleware = None


class AppConfig(AppHandler):
    """
    AppConfig.

    Class for Configuration of aiohttp SubApps
    """

    template: str = "templates"
    path: Path = None
    _middleware = None

    def __init__(self, *args: List, **kwargs: dict):
        self._name = type(self).__name__
        super(AppConfig, self).__init__(*args, **kwargs)
        self.path = APP_DIR.joinpath(self._name)
        # configure templating:
        if self.template:
            template_dir = os.path.join(self.path, self.template)
            # template_dir = self.path.resolve().joinpath(self.template)
            aiohttp_jinja2.setup(self.app, loader=jinja2.FileSystemLoader(template_dir))
        # set the setup_routes
        self.setup_routes()

    async def on_cleanup(self, app):
        await app["redis"].close()

    async def on_startup(self, app):
        kwargs = {"server_settings": {"client_min_messages": "notice"}}
        pool = AsyncPool(
            "pg",
            dsn=app["config"]["asyncpg_url"],
            loop=self._loop,
            timeout=360000,
            **kwargs
        )
        await pool.connect()
        await self.open_connection(pool, app)
        app["pool"] = pool
        # redis Pool
        rd = redis(dsn=app["config"]["cache_url"], loop=self._loop)
        await rd.connection()
        app["redis"] = rd

    async def open_connection(self, pool, app):
        try:
            conn = await pool.acquire()
            if conn:
                connection = conn.engine()
                # print(connection.get_server_version())
                await connection.add_listener(self._name, self.listener)
                await connection.execute(
                    "NOTIFY \"{}\", '= Starting Navigator Notify System = '".format(
                        self._name
                    )
                )
        finally:
            app["connection"] = conn

    async def on_shutdown(self, app):
        await self.close_connection(app["connection"])
        await app["pool"].wait_close(gracefully=False)

    def listener(conn, pid, channel, payload, *args):
        print("Notification from {}: {}, {}".format(channel, payload, args))

    async def close_connection(self, conn):
        try:
            if conn:
                await conn.engine().remove_listener(self._name, self.listener)
                await conn.close()
        except Exception as err:
            print("Error closing Interface connection {}".format(err))
        # finally:
        #     print('= Closing {} connections'.format(self._name))

    def setup_routes(self):
        # set the urls
        # TODO: automatic module loader
        try:
            cls = importlib.import_module(
                "{}.{}".format("apps.{}".format(self._name), "urls"), package="apps"
            )
            routes = getattr(cls, "urls")
        except ImportError as err:
            print(err)
            return False
        for route in routes:
            if inspect.isclass(route.handler) and issubclass(
                route.handler, AbstractView
            ):
                if not route.method:
                    r = self.app.router.add_view(
                        route.url, route.handler, name=route.name
                    )
                elif route.method == "*":
                    r = self.app.router.add_route(
                        "*", route.url, route.handler, name=route.name
                    )
                self.cors.add(r, webview=True)
            elif inspect.isclass(route.handler):
                r = self.app.router.add_view(route.url, route.handler, name=route.name)
                self.cors.add(r, webview=True)
            else:
                if not route.method:
                    r = self.app.router.add_route(
                        "*", route.url, route.handler, name=route.name
                    )
                    self.cors.add(r)
                else:
                    if route.method == "get":
                        r = self.app.router.add_get(
                            route.url, route.handler, name=route.name
                        )
                    elif route.method == "post":
                        r = self.app.router.add_post(
                            route.url, route.handler, name=route.name
                        )
                    elif route.method == "delete":
                        r = self.app.router.add_delete(
                            route.url, route.handler, name=route.name
                        )
                    elif route.method == "patch":
                        r = self.app.router.add_patch(
                            route.url, route.handler, name=route.name
                        )
                    else:
                        raise (
                            "Unsupported Method for Route {}, program: {}".format(
                                route.method, self._name
                            )
                        )
                        return False
                    self.cors.add(r)



import asyncio
import logging
import sys
from pathlib import Path

from aiofile import AIOFile

from . import BaseCommand, CommandError

logger = logging.getLogger("Navigator.creator")
loop = asyncio.get_event_loop()


def read_file(path, filename):
    f = open(path.joinpath("navigator", "commands", "templates", filename), "r")
    return f.read()


def create_dir(dir, name):
    try:
        path = dir.joinpath(name)
        path.mkdir(parents=True, exist_ok=True)
    except FileExistsError as exc:
        pass


def save_file(dir, filename, content):
    async def main(filename, content):
        if content:
            path = dir.joinpath(Path(filename).resolve())
            async with AIOFile(path, "w+") as afp:
                await afp.write(content)
                await afp.fsync()
            return True
        else:
            return False

    return loop.run_until_complete(main(filename, content))


class EnvCommand(BaseCommand):
    help = "Enviroment Commands for Navigator"

    def parse_arguments(self, parser):
        parser.add_argument("--enable-notify", type=bool)
        parser.add_argument("--process-services", type=bool)

    def create(self, options, **kwargs):
        """
        Create can used to create a new Enviroment from scratch
        """
        path = Path(kwargs["project_path"]).resolve()

        env = read_file(path, "env.tpl")
        ini = read_file(path, "ini.tpl")
        settings = read_file(path, "settings.tpl")
        localsettings = read_file(path, "localsettings.tpl")
        run = read_file(path, "run.tpl")
        app = read_file(path, "app.tpl")

        output = "Enviroment Done."
        if options.debug:
            self.write(":: Creating a New Navigator Enviroment", level="INFO")
            self.write("= wait a few minutes", level="WARN")
        # apps, etc, env, services, settings, static/images/js/css, templates
        self.write("* First Step: Creating Directory structure")
        create_dir(path, "apps")
        create_dir(path, "env/testing")
        create_dir(path, "etc")
        create_dir(path, "services")
        create_dir(path, "settings")
        create_dir(path, "static/images")
        create_dir(path, "static/js")
        create_dir(path, "static/css")
        create_dir(path, "templates")
        self.write("* Second Step: Creation of Empty .env File")
        save_file(path, "env/.env", env)
        save_file(path, "env/testing/.env", env)
        save_file(path, "etc/navigator.ini", ini)
        self.write("* Third Step: Creation of Empty settings.py File")
        save_file(path, "settings/settings.py", settings)
        save_file(path, "settings/local_settings.py.example", localsettings)
        self.write("* Fourt Step: Creation of a run.py File")
        save_file(path, "run.py", run)
        save_file(path, "app.py", app)
        self.write("* Fifth Step: adding a *home* template")
        home = read_file(path, "home.html")
        css = read_file(path, "styles.css")
        js = read_file(path, "scripts.js")
        save_file(path, "templates/home.html", home)
        save_file(path, "static/css/styles.css", css)
        save_file(path, "static/js/scripts.js", js)
        return output



import asyncio
from functools import wraps
from typing import Callable

from aiohttp import web
from aiohttp_session import get_session


def authorization_required(func: Callable) -> Callable:
    @wraps(func)
    async def _wrap(*args, **kwargs):
        # handling functions and class methods
        handler = args[0]
        if isinstance(handler, web.Request):
            request = handler
        else:
            request = args[1]
        app = request.app
        router = app.router
        auth = app["auth"]
        for backend in auth.backends():
            if await backend.check_authorization(request):
                return await func(*args, **kwargs)
            else:
                continue
        return web.HTTPFound(router["login"].url_for())

    return _wrap


def login_required(func: Callable) -> Callable:
    func.__login_required__ = True
    return func



#!/usr/bin/env python
import importlib
import os
import sys
import traceback
from argparse import SUPPRESS, ArgumentParser, HelpFormatter
from inspect import signature
from io import TextIOBase
from typing import Any, Callable, Dict, List

from navigator import get_version
from navigator.conf import INSTALLED_APPS


class colors:
    """
    Colors class.

       reset all colors with colors.reset;
       Use as colors.subclass.colorname.
    i.e. colors.fg.red or colors.fg.greenalso, the generic bold, disable,
    underline, reverse, strike through,
    and invisible work with the main class i.e. colors.bold
    """

    reset = "\033[0m"
    bold = "\033[01m"
    disable = "\033[02m"
    underline = "\033[04m"
    reverse = "\033[07m"
    strikethrough = "\033[09m"
    invisible = "\033[08m"

    class fg:
        """
        colors.fg.

        Foreground Color subClass
        """

        black = "\033[30m"
        red = "\033[31m"
        green = "\033[32m"
        orange = "\033[33m"
        blue = "\033[34m"
        purple = "\033[35m"
        cyan = "\033[36m"
        lightgrey = "\033[37m"
        darkgrey = "\033[90m"
        lightred = "\033[91m"
        lightgreen = "\033[92m"
        yellow = "\033[93m"
        lightblue = "\033[94m"
        pink = "\033[95m"
        lightcyan = "\033[96m"


def cPrint(msg, color=None, level="INFO"):
    try:
        if color is not None:
            coloring = colors.bold + getattr(colors.fg, color)
        elif level:
            if level == "INFO":
                coloring = colors.bold + colors.fg.green
            elif level == "SUCCESS":
                coloring = colors.bold + colors.fg.lightgreen
            elif level == "NOTICE":
                coloring = colors.fg.blue
            elif level == "DEBUG":
                coloring = colors.fg.lightblue
            elif level == "WARN":
                coloring = colors.bold + colors.fg.yellow
            elif level == "ERROR":
                coloring = colors.fg.lightred
            elif level == "CRITICAL":
                coloring = colors.bold + colors.fg.red
        else:
            coloring = colors.reset
    except Exception as err:
        print("Wrong color schema {}, error: {}".format(color, str(err)))
        coloring = colors.reset
    print(coloring + msg, colors.reset)


class CommandError(Exception):
    """
    Exception Base Class for raise problems in the execution of a Command
    """

    pass


class CommandNotFound(Exception):
    """
    Exception Base Class for raise problems in the execution of a Command
    """

    pass


class BaseCommand(object):
    parser: Callable
    args: List = []
    action: str = ""
    help: str = "Base Help Command"
    epilog: str = ""

    def __init__(self, args):
        self.args = args
        self.parser = ArgumentParser(
            description=self.help,
            epilog=self.epilog if self.epilog else self.help,
            add_help=False,
        )
        self.parser.add_argument(
            "-v", "--version", action="version", version="%(prog)s 1.0"
        )
        self.parser.add_argument(
            "-h", "--help", action="help", default=SUPPRESS, help="Display this Message"
        )
        self.parser.add_argument(
            "-d", "--debug", action="store_true", help="Enable Debug"
        )
        self.parser.add_argument(
            "--traceback",
            action="store_true",
            help="Return the Traceback on CommandError",
        )
        # get action:
        self.action = self.args.pop(0)
        self.parse_arguments(self.parser)

    def write(self, message, level="INFO"):
        if message:
            cPrint(message, level=level)

    def parse_arguments(self, parser):
        """
        parse_arguments.
            allow for subclassed comands to add custom arguments
        """
        pass

    def get_version(self):
        """
        get_version
            Return the current Navigator Version
        """
        return "Navigator: v.{}".format(get_version())

    def handle(self, **kwargs):
        output: str = ""
        try:

            # calling the internal function:
            if not hasattr(self, self.action):
                self.write(
                    "Error: Method **{}** not found on {}".format(
                        self.action, str(self)
                    ),
                    level="ERROR",
                )
                raise CommandNotFound(
                    "Method {} from {} not Found".format(self, self.action)
                )
            fn = getattr(self, self.action)
            # adding an epilog using the docstring
            self.parser.epilog = str(fn.__doc__)
            # parsing current arguments
            options = self.parser.parse_args(self.args)
            if options.debug:
                self.write("Executing : {}".format(self.action), level="DEBUG")
            sig = signature(fn)
            try:
                if len(sig.parameters) > 0:
                    # send parameters to method
                    output = fn(options, **kwargs)
                else:
                    output = fn()
            except Exception as err:
                if options.traceback:
                    print(traceback.format_exc())
                raise CommandError(
                    "Error Calling Method: {}, error: {}".format(self.action, err)
                )
        except Exception as err:
            if options.traceback:
                print(traceback.format_exc())
            raise CommandError("Error Parsing arguments: {}".format(err))
        finally:
            self.write(output, level="INFO")
            return output


def run_command(**kwargs):
    """
    Running a command in Navigator Enviroment
    """
    if len(sys.argv) > 1:
        args = sys.argv
        script = args.pop(0)
        command = args.pop(0)
        if command is not None:
            # if command is a program, the behavior is different:
            program = "apps.{}".format(command)
            if program in INSTALLED_APPS:
                # is a program command
                cmd = args.pop(0)
                if not args:
                    args.append(cmd)
                clsCommand = "{}Command".format(cmd.capitalize())
                classpath = "{program}.commands.{provider}".format(
                    program=program, provider=clsCommand
                )
                try:
                    module = importlib.import_module(classpath, package=clsCommand)
                    cls = getattr(module, clsCommand)
                except ImportError:
                    raise CommandNotFound(
                        "Command %s for program %s was not found"
                        % (clsCommand, program)
                    )
            else:
                clsCommand = "{}Command".format(command.capitalize())
                # check if is a Navigator Command
                try:
                    classpath = "navigator.commands.{provider}".format(provider=command)
                    module = importlib.import_module(classpath, package="commands")
                    cls = getattr(module, clsCommand)
                except ImportError:
                    raise CommandNotFound(
                        "Command %s was not found o program doesnt exists" % clsCommand
                    )
            try:
                cmd = cls(args)
                output = cmd.handle(**kwargs)
            except Exception as err:
                print(err)



import asyncio
import base64
import time
from datetime import datetime, timedelta

import aioredis
from aiohttp import web

# aiohttp session
from aiohttp_session import get_session, new_session
from aiohttp_session import setup as setup_session
from aiohttp_session.cookie_storage import EncryptedCookieStorage
from aiohttp_session.redis_storage import RedisStorage

# Storages
from cryptography import fernet

from navigator.conf import config
from navigator.libs.modules import AbstractHandler

from .backends import *
from .decorators import login_required


class AuthHandler(AbstractHandler):
    _template = """
        <!doctype html>
            <head></head>
            <body>
                <p>{message}</p>
                <form action="/login" method="POST">
                  Login:
                  <input type="text" name="login">
                  Password:
                  <input type="password" name="password">
                  <input type="submit" value="Login">
                </form>
                <a href="/logout">Logout</a>
            </body>
    """
    _session_type = None
    _pool = None
    _backends = []

    def __init__(
        self,
        type: str = "redis",
        name: str = "AIOHTTP_SESSION",
        backends: list = [],
        **kwargs
    ):
        async def make_redis_pool(url, **kwargs):
            return await aioredis.create_pool(url, **kwargs)

        if type == "redis":
            try:
                url = kwargs["url"]
                del kwargs["url"]
            except KeyError:
                raise Exception("Error: For Redis Storage, you need session URL")
            self._pool = asyncio.get_event_loop().run_until_complete(
                make_redis_pool(url, **kwargs)
            )
            self._session_type = RedisStorage(self._pool, cookie_name=name)
        elif type == "cookie":
            try:
                secret_key = kwargs["secret_key"]
            except KeyError:
                fernet_key = fernet.Fernet.generate_key()
                secret_key = base64.urlsafe_b64decode(fernet_key)
            self._session_type = EncryptedCookieStorage(secret_key, cookie_name=name)
        # TODO: type memcached
        # configure backends
        for backend in backends:
            if backend == "hosts":
                obj = auth_hosts()
            elif backend == "session":
                obj = auth_users()
            self._backends.append(obj)

    def backends(self):
        return self._backends

    def pool(self):
        return self._pool

    async def close(self):
        if self._pool:
            self._pool.close()
            await self._pool.wait_closed()

    async def login_page(self, request):
        username = None
        # check if authorized, instead, return to login
        session = await get_session(request)
        try:
            username = session["username"]
        except KeyError:
            template = self._template.format(message="You need to login")
        if username:
            template = self._template.format(
                message="Hello, {username}!".format(username=username)
            )
        else:
            template = self._template.format(message="You need to login")
        return web.Response(text=template, content_type="text/html")

    async def check_credentials(self, login: str = None, password: str = None):
        if login == "phenobarbital":
            return True
        return False

    async def create_session(self, request, login):
        app = request.app
        session = await new_session(request)
        session["user_id"] = login
        session["username"] = login
        last_visit = session["last_visit"] if "last_visit" in session else "Never"
        session["last_visit"] = time.time()
        session["last_visited"] = "Last visited: {}".format(last_visit)
        app["user"] = session

    async def login(self, request) -> web.Response:
        response = web.HTTPFound("/")
        form = await request.post()
        login = form.get("login")
        password = form.get("password")
        if await self.check_credentials(login, password):
            await self.create_session(request, login)
            raise response
        else:
            template = self._template.format(
                message="Invalid username/password combination"
            )
        raise web.HTTPUnauthorized(text=template, content_type="text/html")

    async def forget_user(self, request):
        session = await get_session(request)
        session.invalidate()
        try:
            app = request.app
            app["user"] = None
            request.user = None
        except Exception as err:
            print(err)

    async def logout(self, request: web.Request) -> web.Response:
        await self.forget_user(request)
        raise web.HTTPSeeOther(location="/")

    async def internal_page(self, request: web.Request) -> web.Response:
        # await check_permission(request, 'public')
        response = web.Response(body="This page is visible for all registered users")
        return response

    @login_required
    async def protected_page(self, request: web.Request) -> web.Response:
        # await check_permission(request, 'protected')
        response = web.Response(body="You are on protected page")
        return response

    def configure(self, app: web.Application) -> web.Application:
        # configure session:
        setup_session(app, self._session_type)
        router = app.router
        router.add_route("GET", "/login", self.login_page, name="index_login")
        router.add_route("POST", "/login", self.login, name="login")
        router.add_route("GET", "/logout", self.logout, name="logout")
        router.add_route("GET", "/public", self.internal_page, name="public")
        router.add_route("GET", "/protected", self.protected_page, name="protected")
        return app



import asyncio
from typing import List


async def run_cmd(cmd: List) -> str:
    print(f'Executing: {" ".join(cmd)}')
    process = await asyncio.create_subprocess_exec(*cmd, stdout=asyncio.subprocess.PIPE)
    out, error = await process.communicate()
    if error:
        raise Exception(error)
    return out.decode("utf8")


def escapeString(value):
    v = value if value != "None" else ""
    v = str(v).replace("'", "''")
    v = "'{}'".format(v) if type(v) == str else v
    return v



#!/usr/bin/env python3
from typing import Any, Awaitable, Callable, Dict, List, Optional

from aiohttp import web
from aiohttp.web import middleware


def check_path(path):
    result = True
    for r in ["/login", "logout", "/static/", "/signin", "/signout", "/_debugtoolbar/"]:
        if path.startswith(r):
            result = False
    return result


@middleware
async def basic_middleware(
    request: web.Request, handler: Callable[[web.Request], Awaitable[web.Response]]
):
    resp = await handler(request)
    return resp



#!/usr/bin/env python3
import sys

from aiohttp import web
from app import Main

from navigator import Application

obj = Application(Main, debug=True)

# declaring auto-reload on debug
app = obj.setup_app()

if __name__ == "__main__":
    obj.run()



import json
from typing import Any

from aiohttp import web
from aiohttp.web import middleware


class AbstractSession(object):
    _backend = None
    _parent = None

    def Backend(self, backend: Any = None):
        self._parent = backend
        self._backend = backend.cache()

    async def decode(self, key):
        pass

    async def decode(self, key):
        pass



import asyncio

import uvloop
from aiogmaps import Client

# make asyncio use the event loop provided by uvloop
asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())


async def get_places(client_id: str, secret: str, place_id: str):
    loop = asyncio.get_event_loop()
    async with Client(client_id=client_id, client_secret=secret, loop=loop) as client:
        return await client.place(place_id="ChIJN1t_tDeuEmsRUsoyG83frY4")



from typing import Any, Callable, Dict, List, Optional

import ujson as json
from aiohttp import web
from aiohttp.web import middleware

from navigator.middlewares import check_path

# TODO: Middleware Class to avoid repeat check_path($0)


@middleware
async def django_session(request, handler):
    id = None
    if not check_path(request.path):
        return await handler(request)
    try:
        id = request.headers.get("sessionid", None)
    except Exception as e:
        print(e)
        id = request.headers.get("X-Sessionid", None)
    if id is not None:
        session = None
        try:
            # first: clear session
            session = request.app["session"]
            await session.logout()  # clear existing session
            if not await session.decode(key=id):
                message = {
                    "code": 403,
                    "message": "Invalid Session",
                    "reason": "Unknown Session ID",
                }
                return web.json_response({"error": message}, status=403)
        except Exception as err:
            print("Error Decoding Session: {}, {}".format(err, err.__class__))
            return await handler(request)
        try:
            request["user_id"] = session["user_id"]
            request["session"] = session
        except Exception as err:
            # TODO: response to an auth error
            message = {
                "code": 403,
                "message": "Invalid Session or Authentication Error",
                "reason": str(err),
            }
            return web.json_response({"error": message}, status=403)
        finally:
            return await handler(request)
    else:
        # TODO: authorization
        return await handler(request)



import os
import sys
from pathlib import Path

from .config import navigatorConfig

__all__ = navigatorConfig


def is_virtualenv():
    return hasattr(sys, "real_prefix") or (
        hasattr(sys, "base_prefix") and sys.base_prefix != sys.prefix
    )


# get Project PATH
if is_virtualenv():
    BASE_DIR = Path(sys.prefix).resolve().parent
else:
    BASE_DIR = Path(os.path.abspath(os.path.dirname(__file__))).resolve().parent.parent

# #BASE_DIR = Path(os.path.abspath(os.path.dirname(__file__))).resolve().parent.parent
# BASE_DIR = Path(sys.prefix).resolve().parent
# if not BASE_DIR:


# extensions dir
EXTENSION_DIR = BASE_DIR.joinpath("extensions")
SERVICES_DIR = BASE_DIR.joinpath("services")
config = navigatorConfig(BASE_DIR)
ENV = config.ENV

# Add Path Navigator to Sys path
sys.path.append(str(BASE_DIR))
sys.path.append(str(EXTENSION_DIR))
sys.path.append(str(SERVICES_DIR))



import asyncio
from http import HTTPStatus
from typing import Callable

from aiohttp import web
from aiohttp_session import get_session


async def auth_middleware(app, handler: Callable) -> web.Response:
    async def middleware(request: web.Request):
        require_login = getattr(handler, "__login_required__", False)
        session = await get_session(request)
        request.user = None
        if require_login:
            if not "user_id" in session:
                raise web.HTTPSeeOther(location="/login")
        # requires authorization
        try:
            for backend in app["auth"].backends():
                result = await backend.check_authorization(request)
                if result:
                    request.user = result
                    break  # authorization was correct
        except Exception as err:
            return web.Response(text="Unauthorized", status=HTTPStatus.UNAUTHORIZED)
        return await handler(request)

    return middleware



import base64
from typing import Any

from navigator.conf import SESSION_PREFIX
from navigator.modules.session.session import AbstractSession

try:
    import ujson as json
except ImportError:
    import json


class djangoSession(AbstractSession):
    async def decode(self, key: str = None):
        try:
            result = await self._backend.get("{}:{}".format(SESSION_PREFIX, key))
            if result:
                data = base64.b64decode(result)
                session_data = data.decode("utf-8").split(":", 1)
                self._session_key = key
                self._session_id = session_data[0]
                if session_data:
                    r = json.loads(session_data[1])
                    self._parent.set_result(r)
                    self._parent.id(self._session_id)
                    return True
            else:
                return False
        except Exception as err:
            print(err)
            logging.debug("Decoding Error: {}".format(err))
            return False

    async def encode(self, key, data):
        raise NotImplementedError



#!/usr/bin/env python
"""Navigator
    Web Framework based on aiohttp, with batteries included.
See:
https://github.com/phenobarbital/navigator-api/tree/master
"""
from setuptools import find_packages, setup

setup(packages=find_packages())



# -*- coding: utf-8 -*-
#!/usr/bin/env python3
import base64
import importlib
import logging
import os
import sys
from pathlib import Path
from types import ModuleType
from typing import Any, Dict, List, Tuple

from cryptography import fernet

# Import Config Class
from navconfig import BASE_DIR, EXTENSION_DIR, QUERYSET_REDIS, asyncpg_url, config

"""
Routes
"""
APP_DIR = BASE_DIR.joinpath("apps")
TEMP_DIR = config.get("TEMP_DIR", fallback="/tmp")
FILES_DIR = config.get("ETL_PATH", fallback="/home/ubuntu/symbits/")
NAV_DIR = BASE_DIR.joinpath("navigator")
STATIC_DIR = BASE_DIR.joinpath("static")
SERVICES_DIR = BASE_DIR.joinpath("services")

"""
Security and debugging
"""
# SECURITY WARNING: keep the secret key used in production secret!
PRODUCTION = config.getboolean("PRODUCTION", fallback=True)
fernet_key = fernet.Fernet.generate_key()
SECRET_KEY = base64.urlsafe_b64decode(fernet_key)
HOSTS = [e.strip() for e in list(config.get("HOSTS", fallback="localhost").split(","))]
# Debug
DEBUG = config.getboolean("DEBUG", fallback=True)
LOCAL_DEVELOPMENT = DEBUG == True and sys.argv[0] == "run.py"
USE_SSL = config.getboolean("ssl", "SSL", fallback=False)

"""
Timezone
"""
# Timezone (For parsedate)
# https://dateparser.readthedocs.io/en/latest/#timezone-and-utc-offset
TIMEZONE = config.get("TIMEZONE", "US/Eastern")

"""
Environment
"""
if DEBUG and LOCAL_DEVELOPMENT:
    ENV = "development"
    CSRF_ENABLED = False
    SSL = False
    SSL_VERIFY = False
    SSL_CERT = None
    SSL_KEY = None
    PREFERRED_URL_SCHEME = "http"
    CREDENTIALS_REQUIRED = False
    ENABLE_TOKEN_APP = False
else:
    CREDENTIALS_REQUIRED = True
    ENABLE_TOKEN_APP = True
    if PRODUCTION == False and DEBUG == True:
        ENV = "development"
        CSRF_ENABLED = False
    elif PRODUCTION == True and DEBUG == True:
        ENV = "staging"
        CSRF_ENABLED = True
    else:
        ENV = "production"
        CSRF_ENABLED = True
    try:
        SSL_CERT = config.get("CERT")
        SSL_KEY = config.get("KEY")
        PREFERRED_URL_SCHEME = "https"
    except Exception as e:
        SSL_CERT = None
        SSL_KEY = None
        PREFERRED_URL_SCHEME = "http"

# Temp File Path
files_path = BASE_DIR.joinpath("temp")

"""
Basic Information
"""
EMAIL_CONTACT = config.get("EMAIL_CONTACT", section="info", fallback="foo@example.com")
API_NAME = config.get("API_NAME", section="info", fallback="Navigator")

# get settings
from navconfig.conf import *

#######################
##
## APPS CONFIGURATION
##
#######################
"""
Applications
"""
INSTALLED_APPS: List = []
DATABASES: Dict = {}

if APP_DIR.is_dir():
    for item in APP_DIR.iterdir():
        # for item in os.listdir(APPS_DIR):
        if item.name != "__pycache__":
            if item.is_dir():
                name = item.name
                if not name in INSTALLED_APPS:
                    app_name = "apps.{}".format(item.name)
                    path = APP_DIR.joinpath(name)
                    url_file = path.joinpath("urls.py")
                    try:
                        i = importlib.import_module(app_name, package="apps")
                        if isinstance(i, ModuleType):
                            # is a Navigator Program
                            INSTALLED_APPS += (app_name,)
                    except ImportError as err:
                        print("ERROR: ", err)
                        continue
                    # schema configuration
                    DATABASES[item.name] = {
                        "ENGINE": config.get("DBENGINE"),
                        "NAME": config.get("DBNAME"),
                        "USER": config.get("DBUSER"),
                        "OPTIONS": {
                            "options": "-c search_path=" + item.name + ",troc,public",
                        },
                        #'PARAMS': {
                        #    'readonly': True,
                        # },
                        "SCHEMA": item.name,
                        "PASSWORD": config.get("DBPWD"),
                        "HOST": config.get("DBHOST", fallback="localhost"),
                        "PORT": config.get("DBPORT"),
                    }


"""
Per-Program Settings
"""
program: str
for program in INSTALLED_APPS:
    settings = "apps.{}.settings".format(program)
    try:
        i = importlib.import_module(settings)
        globals()[program] = i
    except ImportError as err:
        pass

"""
Config dict for aiohttp
"""
Context = {
    "DEBUG": DEBUG,
    "DEVELOPMENT": (not PRODUCTION),
    "PRODUCTION": PRODUCTION,
    "SECRET_KEY": SECRET_KEY,
    "env": ENV,
    "DATABASES": DATABASES,
    "asyncpg_url": asyncpg_url,
    "cache_url": QUERYSET_REDIS,
}



#!/usr/bin/env python3
import argparse
import asyncio
import ssl
import sys
import typing

import aiohttp_cors
import uvloop
from aiohttp import web

# make asyncio use the event loop provided by uvloop
asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

import inspect
import logging
import signal
from concurrent.futures import ThreadPoolExecutor
from functools import wraps
from typing import Any, Callable, Optional

# from aiohttp_swagger import setup_swagger
import sockjs
from aiohttp import web
from aiohttp.abc import AbstractView
from aiohttp_session import get_session
from aiohttp_session import setup as setup_session
from aiohttp_session.redis_storage import RedisStorage

# from apps.setup import app_startup
from aiohttp_utils import run as runner

from navigator.applications import AppBase, AppHandler, app_startup
from navigator.conf import (
    APP_DIR,
    BASE_DIR,
    EMAIL_CONTACT,
    INSTALLED_APPS,
    LOCAL_DEVELOPMENT,
    SECRET_KEY,
    SESSION_PREFIX,
    SESSION_STORAGE,
    SESSION_URL,
    STATIC_DIR,
    Context,
    config,
)

# Exception Handlers
from navigator.handlers import nav_exception_handler, shutdown

# get the authentication library
from navigator.modules.auth import AuthHandler, auth_middleware, login_required
from navigator.resources import WebSocket, channel_handler

__version__ = "1.0.0.dev0"


def get_version():
    return __version__


def Response(
    content: Any,
    text: str = "",
    body: Any = None,
    status: int = 200,
    headers: dict = None,
    content_type: str = "text/plain",
    charset: str = "utf-8",
) -> web.Response:
    """
    Response.
    Web Response Definition for Navigator
    """
    response = {"content_type": content_type, "charset": charset, "status": status}
    if headers:
        response["headers"] = headers
    if isinstance(content, str) or text is not None:
        response["text"] = content if content else text
    else:
        response["body"] = content if content else body
    return web.Response(**response)


class Application(object):
    app: Any = None
    _auth = None
    debug = False
    parser = None
    use_ssl = False
    _routes: list = []
    _reload = False
    _logger = None
    path = ""
    host = "0.0.0.0"
    port = 5000
    _loop = None

    def __init__(self, app: AppHandler = None, *args: typing.Any, **kwargs: typing.Any):
        # configuring asyncio loop
        try:
            self._loop = asyncio.get_event_loop()
        except RuntimeError:
            logger.debug(
                "Couldn't get event loop for current thread. Creating a new event loop to be used!"
            )
            self._loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self._loop)
        # self._loop.set_exception_handler(nav_exception_handler)
        # May want to catch other signals too
        signals = (signal.SIGHUP, signal.SIGTERM, signal.SIGINT)
        for s in signals:
            self._loop.add_signal_handler(
                s, lambda s=s: asyncio.create_task(shutdown(self._loop, s))
            )
        self._executor = ThreadPoolExecutor()
        if "debug" in kwargs:
            self.debug = kwargs["debug"]
        self.parser = argparse.ArgumentParser(description="Navigator App")
        self.parser.add_argument("--path")
        self.parser.add_argument("--host")
        self.parser.add_argument("--port")
        self.parser.add_argument("-r", "--reload", action="store_true")
        self.parser.add_argument(
            "-d", "--debug", action="store_true", help="Enable Debug"
        )
        self.parser.add_argument(
            "--traceback", action="store_true", help="Return the Traceback on Error"
        )
        self.parse_arguments()
        if not app:
            # create an instance of AppHandler
            self.app = AppBase(Context)
        else:
            self.app = app(Context)
        self._logger = self.get_logger(self.app.Name)

    def parse_arguments(self):
        args = self.parser.parse_args()
        try:
            self.path = args.path
        except (KeyError, ValueError, TypeError):
            self.path = None
        try:
            self.host = args.host
            if not self.host:
                self.host = "0.0.0.0"
        except (KeyError, ValueError, TypeError):
            self.host = "0.0.0.0"
        try:
            self.port = args.port
            if not self.port:
                self.port = 5000
        except (KeyError, ValueError, TypeError):
            self.port = 5000
        try:
            self.debug = args.debug
        except (KeyError, ValueError, TypeError):
            pass
        try:
            self._reload = args.reload
        except (KeyError, ValueError, TypeError):
            self._reload = False

    def get_app(self) -> web.Application:
        return self.app.App

    def get_logger(self, name="Navigator"):
        logging_format = f"[%(asctime)s] %(levelname)-5s %(name)-{len(name)}s "
        # logging_format += "%(module)-7s::l%(lineno)d: "
        # logging_format += "%(module)-7s: "
        logging_format += "%(message)s"
        logging.basicConfig(
            format=logging_format, level=logging.INFO, datefmt="%Y:%m:%d %H:%M:%S"
        )
        logging.getLogger("asyncio").setLevel(logging.INFO)
        logging.getLogger("websockets").setLevel(logging.INFO)
        logging.getLogger("aiohttp.web").setLevel(logging.INFO)
        return logging.getLogger(name)

    def setup_app(self) -> web.Application:
        app = self.get_app()
        # # TODO: iterate over modules folder
        self._auth = AuthHandler(
            type=SESSION_STORAGE,
            name=SESSION_PREFIX,
            url=SESSION_URL,
            backends=(
                "session",
                "hosts",
            ),
        )
        self._auth.configure(app)
        # adding middleware for Authorization
        app.middlewares.append(auth_middleware)
        # setup The Application and Sub-Applications Startup
        app_startup(INSTALLED_APPS, app, Context)
        app["auth"] = self._auth
        # Configure Routes
        self.app.configure()
        cors = aiohttp_cors.setup(
            app,
            defaults={
                "*": aiohttp_cors.ResourceOptions(
                    allow_credentials=True,
                    expose_headers="*",
                    allow_methods="*",
                    allow_headers="*",
                    max_age=3600,
                )
            },
        )
        self.app.setup_cors(cors)
        self.app.setup_docs()
        # # auto-configure swagger
        # long_description = """
        # Asynchronous RESTful API for data source connections, REST consumer and Query API, used by Navigator, powered by MobileInsight
        # """
        # setup_swagger(app,
        #     api_base_url='/',
        #     title='API',
        #     api_version='2.0.0',
        #     description=long_description,
        #     contact=EMAIL_CONTACT,
        #     swagger_url="/api/doc")
        # TODO: configure documentation
        return app

    def add_websockets(self) -> None:
        """
        add_websockets.
        description: enable support for websockets in the main App
        """
        app = self.get_app()
        if self.debug:
            print("Enabling WebSockets")
        # websockets
        app.router.add_route("GET", "/ws", WebSocket)
        # websocket channels
        app.router.add_route("GET", "/ws/{channel}", channel_handler)

    def add_routes(self, routes: list) -> None:
        """
        add_route
        description: append a route to routes dict
        """
        # TODO: avoid to add same route different times
        # self._routes.append(route)
        self.get_app().add_routes(routes)

    def add_static(self, route: str, path: str):
        """
        add_static
        description: register new route to static path.
        """
        self.get_app().add_static(route, path)

    def add_view(self, route: str, handler: Any):
        self.get_app().router.add_view(route, handler)

    def threaded_func(self, func: Callable, threaded: bool = False):
        @wraps(func)
        async def _wrap(request):
            result = None
            try:
                if threaded:

                    def blocking_function():
                        return asyncio.new_event_loop().run_until_complete(
                            func(request)
                        )

                    result = await self._loop.run_in_executor(
                        self._executor, blocking_function
                    )
                else:
                    result = await func(request)
                return result
            except Exception as err:
                self._logger.exception(err)

        return _wrap

    def route(self, route: str, method: str = "GET", threaded: bool = False):
        """
        route.
        description: decorator for register a new HTTP route.
        """

        def _decorator(func):
            self.app.App.router.add_route(
                method, route, self.threaded_func(func, threaded)
            )
            return func

        return _decorator

    def add_get(self, route: str, threaded: bool = False) -> Callable:
        def _decorator(func):
            self.app.App.router.add_get(route, self.threaded_func(func, threaded))
            return func

        return _decorator

    def Response(self, content: Any) -> web.Response:
        return web.Response(text=content)

    def get(self, route: str):
        def _decorator(func):
            self.app.App.router.add_get(route, func)

            @wraps(func)
            async def _wrap(request, *args, **kwargs):
                try:
                    return f"{func(request, args, **kwargs)}"
                except Exception as err:
                    self._logger.exception(err)

            return _wrap

        return _decorator

    def add_sock_endpoint(
        self, handler: Callable, name: str, route: str = "/sockjs/"
    ) -> None:
        app = self.get_app()
        sockjs.add_endpoint(app, handler, name=name, prefix=route)

    def run(self):
        # getting the app
        # getting the resource App
        app = self.setup_app()
        if self.debug is True:
            if LOCAL_DEVELOPMENT:
                import aiohttp_debugtoolbar
                from aiohttp_debugtoolbar import toolbar_middleware_factory

                aiohttp_debugtoolbar.setup(app, hosts=[self.host], enabled=True)
            if self._reload:
                runner(
                    app=app,
                    app_uri="navigator.runserver:app",
                    reload=True,
                    port=self.port,
                    host=self.host,
                )
            else:
                web.run_app(app, host=self.host, port=self.port)
        else:
            if self.use_ssl:
                ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
                ssl_context.load_cert_chain(SSL_CERT, SSL_KEY)
                web.run_app(
                    app, host=self.host, port=self.port, ssl_context=ssl_context
                )
            if self.path:
                web.run_app(app, path=self.path)
            else:
                web.run_app(app, host=self.host, port=self.port)



import asyncio
import datetime
import inspect
import json
import logging
import traceback
from abc import ABC, ABCMeta, abstractmethod, abstractproperty
from functools import partial
from logging.config import dictConfig
from typing import Any, Callable, Dict, List, Optional
from urllib import parse

import aiohttp_cors
from aiohttp import web
from aiohttp.abc import AbstractView
from aiohttp.web import Response, StreamResponse
from aiohttp.web_exceptions import (
    HTTPClientError,
    HTTPInternalServerError,
    HTTPMethodNotAllowed,
    HTTPNoContent,
    HTTPNotImplemented,
    HTTPUnauthorized,
)
from aiohttp_cors import CorsViewMixin
from asyncdb.providers.memcache import memcache
from settings.settings import MEMCACHE_HOST, MEMCACHE_PORT, logging_config, loglevel

from navigator.libs.encoders import DefaultEncoder

dictConfig(logging_config)


class BaseHandler(CorsViewMixin):
    _config = None
    _mem = None
    _now = None
    _loop = None
    logger: logging.Logger
    _lasterr = None

    cors_config = {
        "*": aiohttp_cors.ResourceOptions(
            allow_credentials=True,
            allow_headers="*",
        )
    }

    def __init__(self, *args, **kwargs):
        CorsViewMixin.__init__(self)
        self._now = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
        self._loop = asyncio.get_event_loop()
        self.logger = logging.getLogger("Navigator")
        self.logger.setLevel(loglevel)
        self.post_init(self, *args, **kwargs)

    def post_init(self, *args, **kwargs):
        pass

    def now(self):
        return self._now

    def log(self, message: str):
        self.logger.info(message)

    # function returns
    def no_content(
        self, request, headers, content_type: str = "application/json"
    ) -> web.Response:
        response = HTTPNoContent(content_type=content_type)
        response.headers["Pragma"] = "no-cache"
        for header, value in headers.items():
            response.headers[header] = value
        return response

    def response(
        self,
        request: web.Request,
        response: str = "",
        state: int = 200,
        headers: dict = {},
        **kwargs
    ) -> web.Response:
        args = {"status": state, "content_type": "application/json", **kwargs}
        if isinstance(response, dict):
            args["text"] = json.dumps(response, cls=DefaultEncoder)
        else:
            args["body"] = response
        obj = web.Response(**args)
        for header, value in headers.items():
            obj.headers[header] = value
        return obj

    def json_response(self, response={}, headers={}, state=200, encoder=None):
        if encoder is not None:
            if inspect.isclass(encoder):
                # its a class-based Encoder
                jsonfn = partial(json.dumps, cls=encoder)
            else:
                # its a function
                jsonfn = partial(json.dumps, default=encoder)
        else:
            jsonfn = json.dumps
        obj = web.json_response(response, status=state, dumps=jsonfn)
        for header, value in headers.items():
            obj.headers[header] = value
        return obj

    def critical(
        self,
        request: web.Request,
        exception: Exception = None,
        traceback=None,
        state: int = 500,
        headers: dict = {},
        **kwargs
    ) -> web.Response:
        # TODO: process the exception object
        response_obj = {
            "status": "Failed",
            "reason": str(exception),
            "stacktrace": traceback,
        }
        args = {
            "text": json.dumps(response_obj),
            "reason": "Server Error",
            "content_type": "application/json",
            **kwargs,
        }
        if state == 500:  # bad request
            obj = HTTPInternalServerError(**args)
        else:
            obj = HTTPServerError(**args)
        for header, value in headers.items():
            obj.headers[header] = value
        return obj

    def error(
        self,
        request: dict,
        response: dict = {},
        exception: Exception = None,
        state: int = 400,
        headers: dict = {},
        **kwargs
    ) -> web.Response:
        # TODO: process the exception object
        response_obj = {"status": "Failed"}
        if exception:
            response_obj["reason"] = str(exception)
        args = {**kwargs}
        if isinstance(response, dict):
            response_obj = {**response_obj, **response}
            args["content_type"] = "application/json"
            args["text"] = json.dumps(response_obj, cls=DefaultEncoder)
        else:
            args["body"] = response
        # defining the error
        if state == 400:  # bad request
            obj = web.HTTPBadRequest(**args)
        elif state == 401:  # unauthorized
            obj = web.HTTPUnauthorized(**args)
        elif state == 403:  # forbidden
            obj = web.HTTPForbidden(**args)
        elif state == 404:  # not found
            obj = web.HTTPNotFound(**args)
        else:
            obj = web.HTTPClientError(**args)
        for header, value in headers.items():
            obj.headers[header] = value
        return obj

    def not_implemented(
        self, request: web.Request, response: dict = {}, headers: dict = {}, **kwargs
    ) -> web.Response:
        args = {
            "text": json.dumps(response),
            "reason": "Method not Implemented",
            "content_type": "application/json",
            **kwargs,
        }
        response = HTTPNotImplemented(**args)
        for header, value in headers.items():
            response.headers[header] = value
        return response

    def not_allowed(
        self,
        request: web.Request,
        response: dict = {},
        headers: dict = {},
        allowed: dict = {},
        **kwargs
    ) -> web.Response:
        if not allowed:
            allow = self._allowed
        else:
            allow = allowed
        args = {
            "method": request.method,
            "text": json.dumps(response),
            "reason": "Method not Allowed",
            "content_type": "application/json",
            "allowed_methods": allow,
            **kwargs,
        }
        if allowed:
            headers["Allow"] = ",".join(allow)
        else:
            headers["Allow"] = ",".join(allow)
        obj = HTTPMethodNotAllowed(**args)
        for header, value in headers.items():
            obj.headers[header] = value
        return obj

    def query_parameters(self, request: web.Request) -> dict:
        return {key: val for (key, val) in request.query.items()}

    async def body(self, request: web.Request) -> str:
        body = None
        try:
            if request.body_exists:
                body = await request.read()
                body = body.decode("ascii")
        except Exception as e:
            pass
        finally:
            return body

    async def json_data(self, request):
        return await request.json()


class BaseView(web.View, BaseHandler, AbstractView):
    # _allowed: FrozenSet = frozenset()  # {'get', 'post', 'put', 'patch', 'delete', 'option'}

    def __init__(self, request, *args, **kwargs):
        AbstractView.__init__(self, request)
        BaseHandler.__init__(self, *args, **kwargs)
        self._request = request

    def get_args(self) -> dict:
        args = {}
        for arg in self.request.match_info:
            try:
                val = self.request.match_info.get(arg)
                object.__setattr__(self, arg, val)
                args[arg] = val
            except AttributeError:
                pass
        qry = {}
        try:
            qry = {key: val for (key, val) in self.request.rel_url.query.items()}
        except Exception as err:
            print(err)
            pass
        args = {**args, **qry}
        return args

    async def post_data(self) -> dict:
        args = {}
        if self.request.headers.get("Content-Type") == "application/json":
            return await self.request.json()
        try:
            args = await self.request.post()
            if not args or len(args) == 0:
                if self.request.body_exists:
                    body = await self.request.read()
                    body = body.decode("ascii")
                    if body:
                        try:
                            args = dict(
                                (k, v if len(v) > 1 else v[0])
                                for k, v in parse.parse_qs(body).items()
                            )
                        except (KeyError, ValueError):
                            pass
        finally:
            print(args)
            return args

    def no_content(self, headers: Dict) -> web.Response:
        response = HTTPNoContent(content_type="application/json")
        response.headers["Pragma"] = "no-cache"
        for header, value in headers.items():
            response.headers[header] = value
        return response


class DataView(BaseView):
    _mcache: Any = None
    _connection: Any = None
    _redis: Any = None

    def post_init(self, *args, **kwargs):
        mem_params = {"host": MEMCACHE_HOST, "port": MEMCACHE_PORT}
        self._mcache = memcache(params=mem_params)

    async def connection(self):
        return self._connection

    async def connect(self, request):
        await self._mcache.connection()
        self._connection = await request.app["pool"].acquire()
        self._redis = request.app["redis"]

    async def close(self):
        if self._mcache and self._mcache.is_connected():
            await self._mcache.close()
            self._mcache = None
        if self._connection:
            await self._connection.close()
            self._connection = None

    async def asyncdb(self, request):
        db = None
        try:
            pool = request.app["pool"]
            if pool.is_connected():
                conn = await pool.acquire()
                db = asyncORM(db=conn, loop=self._loop)
                return db
        finally:
            return db

    async def query(self, sql):
        result = None
        if self._connection:
            try:
                result, error = await self._connection.query(sql)
                if error:
                    print(error)
                    result = None
                    self._lasterr = error
            except Exception as err:
                print(err)
                result = None
                self._lasterr = err
            finally:
                return result

    async def queryrow(self, sql):
        result = None
        if self._connection:
            try:
                result, error = await self._connection.queryrow(sql)
                if error:
                    result = None
            except Exception as err:
                print(err)
            finally:
                return result

    async def execute(self, sql):
        result = None
        if self._connection:
            try:
                result, error = await self._connection.execute(sql)
                if error:
                    result = None
            except Exception as err:
                print(err)
            finally:
                return result

    """
    Meta-Operations
    """

    def table(self, table):
        try:
            return self._query_raw.format_map(SafeDict(table=table))
        except Exception as e:
            print(e)
            return False

    def fields(self, sentence, fields=None):
        _sql = False
        if not fields:
            _sql = sentence.format_map(SafeDict(fields="*"))
        elif type(fields) == str:
            _sql = sentence.format_map(SafeDict(fields=fields))
        elif type(fields) == list:
            _sql = sentence.format_map(SafeDict(fields=",".join(fields)))
        return _sql

    """
    where
      add WHERE conditions to SQL
    """

    def where(self, sentence, where):
        sql = ""
        if sentence:
            where_string = ""
            if not where:
                sql = sentence.format_map(SafeDict(where_cond=""))
            elif type(where) == dict:
                where_cond = []
                for key, value in where.items():
                    # print("KEY {}, VAL: {}".format(key, value))
                    if type(value) == str or type(value) == int:
                        if value == "null" or value == "NULL":
                            where_string.append("%s IS NULL" % (key))
                        elif value == "!null" or value == "!NULL":
                            where_string.append("%s IS NOT NULL" % (key))
                        elif key.endswith("!"):
                            where_cond.append("%s != %s" % (key[:-1], value))
                        else:
                            if (
                                type(value) == str
                                and value.startswith("'")
                                and value.endswith("'")
                            ):
                                where_cond.append("%s = %s" % (key, "{}".format(value)))
                            elif type(value) == int:
                                where_cond.append("%s = %s" % (key, "{}".format(value)))
                            else:
                                where_cond.append(
                                    "%s = %s" % (key, "'{}'".format(value))
                                )
                    elif type(value) == bool:
                        val = str(value)
                        where_cond.append("%s = %s" % (key, val))
                    else:
                        val = ",".join(map(str, value))
                        if type(val) == str and "'" not in val:
                            where_cond.append("%s IN (%s)" % (key, "'{}'".format(val)))
                        else:
                            where_cond.append("%s IN (%s)" % (key, val))
                # if 'WHERE ' in sentence:
                #    where_string = ' AND %s' % (' AND '.join(where_cond))
                # else:
                where_string = " WHERE %s" % (" AND ".join(where_cond))
                print("WHERE cond is %s" % where_string)
                sql = sentence.format_map(SafeDict(where_cond=where_string))
            elif type(where) == str:
                where_string = where
                if not where.startswith("WHERE"):
                    where_string = " WHERE %s" % where
                sql = sentence.format_map(SafeDict(where_cond=where_string))
            else:
                sql = sentence.format_map(SafeDict(where_cond=""))
            del where
            del where_string
            return sql
        else:
            return False

    def limit(self, sentence, limit=1):
        """
        LIMIT
          add limiting to SQL
        """
        if sentence:
            return "{q} LIMIT {limit}".format(q=sentence, limit=limit)
        return self

    def orderby(self, sentence, ordering=[]):
        """
        LIMIT
          add limiting to SQL
        """
        if sentence:
            if type(ordering) == str:
                return "{q} ORDER BY {ordering}".format(q=sentence, ordering=ordering)
            elif type(ordering) == list:
                return "{q} ORDER BY {ordering}".format(
                    q=sentence, ordering=", ".join(ordering)
                )
        return self



import asyncio

from aiohttp import web
from aiohttp_session import get_session

from navigator.conf import HOSTS


class auth_hosts(object):
    """
    BasicHosts.
       Use basic Host authorization, simply creating a list of allowed hosts
    """

    async def check_authorization(self, request: web.Request) -> bool:
        if request.host in HOSTS:
            return {"host": request.host, "user_id": request.host}
        try:
            if request.headers["origin"] in HOSTS:
                return {"host": request.headers["origin"], "user_id": request.host}
        except KeyError:
            return False
        return False


class auth_users(object):
    async def check_authorization(self, request: web.Request) -> bool:
        session = await get_session(request)
        if "user_id" not in session:
            return False
        return session


class auth_django(object):
    async def check_authorization(self, request: web.Request) -> bool:
        session = await get_session(request)
        if "user_id" not in session:
            return False
        return session
